<?php

namespace Drp\UserManagement;

use Carbon\Carbon;
use Drp\UserManagement\Contracts\ManagedUser;
use Drp\UserManagement\Events\AccountActivated;
use Illuminate\Auth\Authenticatable;
use Illuminate\Auth\Passwords\CanResetPassword;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Foundation\Auth\Access\Authorizable;
use Illuminate\Notifications\Notifiable;

/**
 * Class User
 * @package Drp\UserManagement
 * @SuppressWarnings(PHPMD.TooManyPublicMethods)
 */
class User extends Model implements ManagedUser
{
    use Authenticatable, CanResetPassword, Notifiable, Authorizable;

    /**
     * The attributes that should be cast to native types.
     *
     * @var array
     */
    protected $casts = [
        'approved' => 'boolean',
    ];

    /**
     * The attributes that should be mutated to dates.
     *
     * @var array
     */
    protected $dates = [
        'login_link_expiry',
    ];

    /**
     * The database table used by the model.
     *
     * @var string
     */
    protected $table = 'users';

    /**
     * The attributes that are mass assignable.
     *
     * @var array
     */
    protected $fillable = [
        'name',
        'email',
        'role',
        'password',
        'approved',
        'login_link_expiry',
        'first_login_token'
    ];

    /**
     * Get array of users that meet the given roles.
     *
     * @param string|array $roles
     * @return \Illuminate\Database\Eloquent\Collection
     */
    public function usersByRoles($roles)
    {
        if (!is_array($roles)) {
            $roles = [$roles];
        }

        return $this->whereIn('role', $roles)->get();
    }

    /**
     * Return all the admin users.
     *
     * @return \Illuminate\Database\Eloquent\Collection
     */
    public function allAdmins()
    {
        return $this->usersByRoles(UserRoles::adminRoles());
    }

    /**
     * Return all the super admin users.
     *
     * @return \Illuminate\Database\Eloquent\Collection
     */
    public function allSuperAdmins()
    {
        return $this->usersByRoles(UserRoles::superAdminRoles());
    }

    /**
     * Return if the user has admin access.
     *
     * @return boolean
     */
    public function hasAdminAccess()
    {
        return in_array($this->role, UserRoles::adminRoles());
    }

    /**
     * Determine if the user has super user access.
     *
     * @return boolean
     */
    public function hasSuperAdminAccess()
    {
        return in_array($this->role, UserRoles::superAdminRoles());
    }


    /**
     * Determine if the user can manage other admins.
     *
     * @return boolean
     */
    public function canManageAdmins()
    {
        return $this->hasSuperAdminAccess();
    }

    /**
     * Return if the user has the given role.
     *
     * @param string $role The Role to check against
     * @return boolean
     */
    public function hasRole($role)
    {
        return $this->role === $role;
    }

    /**
     * Return if the user has and of the given roles.
     *
     * @param string|array $roles The roles to check
     * @param mixed ...$additionalRoles Unlimited OPTIONAL number of additional roles to check
     * @return boolean
     */
    public function hasAnyRole($roles, ...$additionalRoles)
    {
        return in_array($this->role, array_merge(array_wrap($roles), $additionalRoles));
    }

    /**
     * Finds a user by their one time login token.
     *
     * @param string $token The token to retrieve the user
     * @return mixed
     * @throws \Illuminate\Database\Eloquent\ModelNotFoundException If no user was found.
     */
    public function getByFirstLoginToken($token)
    {
        return $this->where('first_login_token', $token)->firstOrFail();
    }

    /**
     * @return \Illuminate\Support\Collection
     */
    public function allAwaitingApproval()
    {
        return static::where('approved', false)->get();
    }

    /**
     * Return if the user is awaiting approval.
     *
     * @return boolean
     */
    public function requiresApproving()
    {
        return !$this->approved;
    }

    /**
     * Approve the user.
     *
     * @return void
     */
    public function approve()
    {
        $this->approved = true;
        $this->save();

        event(new AccountActivated($this));
    }

    /**
     * Removes a one time login token from a user.
     *
     * @return static
     */
    public function removeFirstLoginToken()
    {
        $this->first_login_token = null;

        $this->save();

        return $this;
    }

    /**
     * Store the login token against the user.
     *
     * @param mixed $token
     * @param \DateTime $expiry
     * @return static
     */
    public function addLoginToken($token, \DateTime $expiry)
    {
        $this->first_login_token = $token;
        $this->login_link_expiry = $expiry;
        $this->save();

        return $this;
    }

    /**
     * Return the login token the user can use to access the site.
     *
     * @return string|null
     */
    public function loginToken()
    {
        return $this->first_login_token;
    }

    /**
     * Return the expiry of the login token.
     *
     * @return Carbon|null
     */
    public function loginExpiry()
    {
        return $this->login_link_expiry;
    }
}
